/////////////////////////////////////////////////////////////
//
// image.cpp
//
// Mise en oeuvre des methodes de la classe Image
//
/////////////////////////////////////////////////////////////

#define IMAGE_MAIN


//#include "verbose.hpp"
#include <stdlib.h>
#include <assert.h>

// Methodes privees ////////////////////////////////////////
template <class T>
void Image<T>::Allocate (const int nbRow, const int nbCol)
{
  height       = nbRow;
  width        = nbCol;
  bufferSize   = height * width;
  deleteBuffer = true;
  buffer       = new T[bufferSize];
}


template <class T>
void Image<T>::DeAllocate ()
{
  if (deleteBuffer)
  {
    if (buffer) delete [] buffer;
    height       = 0;
    width        = 0;
    bufferSize   = 0;
    deleteBuffer = false;
    buffer       = 0;
  }
}


// Constructeurs ///////////////////////////////////////////
template <class T>
Image<T>::Image ()
{
  height       = 0;
  width        = 0;
  bufferSize   = 0;
  deleteBuffer = false;
  buffer       = 0;
}


template <class T>
Image<T>::Image (const int nbRow, const int nbCol)
{
#ifndef RELEASE
      assert((nbRow > 0) && (nbCol > 0));
#endif
  Allocate (nbRow, nbCol);
}


template <class T>
Image<T>::Image (const int nbRow, const int nbCol, const T &val)
{
#ifndef RELEASE
  assert((nbRow > 0) && (nbCol > 0));
#endif
  Allocate (nbRow, nbCol);

  for (int i = 0; i < bufferSize; i++)
    buffer[i] = val;
}


template <class T>
Image<T>::Image (const Image &ima)
{
//  #ifndef RELEASE
//    assert(ima.BufferSize() > 0);
//  #endif
  if (ima.BufferSize() > 0) { // pour STL
    Allocate (ima.NbRow(), ima.NbCol());
    for (int i = 0; i < bufferSize; i++)
      buffer[i] = ima[i];
  } else {
    height       = 0;
    width        = 0;
    bufferSize   = 0;
    deleteBuffer = false;
    buffer       = 0;
  }
}



template <class T>
Image<T>::Image (const int nbRow, const int nbCol,  const T *const buf)
{
#ifndef RELEASE
  assert((buf != NULL) && (nbRow > 0) && (nbCol > 0));
#endif
  Allocate (nbRow, nbCol);

  for (int i = 0; i < bufferSize; i++)
    buffer[i] = buf[i];
}


// Destructeur ////////////////////////////////////////////////////////
template <class T>
Image<T>::~Image ()
{
  DeAllocate ();
}



// Methodes d'acces ///////////////////////////////////////////////////
template <class T>
int Image<T>::NbRow () const
{
  return height;
}


template <class T>
int Image<T>::NbCol () const
{
  return width;
}


template <class T>
int Image<T>::BufferSize () const
{
  return bufferSize;
}


// operateurs d'acces /////////////////////////////////////////////////
template <class T>
const T& Image<T>::operator () (const int row, const int col) const
{
#ifndef RELEASE
  assert((row >= 0) && (row < height));
  assert((col >= 0) && (col < width));
#endif
  return buffer[row * width + col];
}


template <class T>
const T& Image<T>::operator [] (const int i) const
{
#ifndef RELEASE
  assert((i >= 0) && (i < bufferSize));
#endif
  return buffer[i];
}


// operateurs de modification /////////////////////////////////////////
template <class T>
T& Image<T>::operator () (const int row, const int col)
{
#ifndef RELEASE
  assert((row >= 0) && (row < height));
  assert((col >= 0) && (col < width));
#endif
  return buffer[row * width + col];
}


template <class T>
T& Image<T>::operator [] (const int i)
{
#ifndef RELEASE
  assert((i >= 0) && (i < bufferSize));
#endif
  return buffer[i];
}


template <class T>
Image<T>& Image<T>::operator = (const Image &ima)
{
//  #ifndef RELEASE
//    assert(ima.BufferSize() != 0);
//  #endif
  DeAllocate ();
  if (ima.BufferSize() > 0) { // pour STL
    Allocate (ima.NbRow(), ima.NbCol());
    for (int i = 0; i < bufferSize; i++)
      buffer[i] = ima[i];
   } else {
     height       = 0;
     width        = 0;
     bufferSize   = 0;
     deleteBuffer = false;
     buffer       = 0;
   }

  return *this;
}


// Methodes de modification ///////////////////////////////////////////
template <class T>
T* Image<T>::PixelPointer (const int row, const int col)
{
#ifndef RELEASE
  assert((row >= 0) && (row < height));
  assert((col >= 0) && (col < width));
#endif
  return buffer + row * width + col;
}

template <class T>
T* Image<T>::PixelPointer (const int i)
{
#ifndef RELEASE
  assert((i >= 0) && (i < bufferSize));
#endif
  return buffer + i;
}


template <class T>
const T* Image<T>::PixelPointer (const int row, const int col) const
{
#ifndef RELEASE
  assert((row >= 0) && (row < height));
  assert((col >= 0) && (col < width));
#endif
  return buffer + row * width + col;
}

template <class T>
const T* Image<T>::PixelPointer (const int i) const
{
#ifndef RELEASE
  assert((i >= 0) && (i < bufferSize));
#endif
  return buffer + i;
}



template <class T>
void Image<T>::InitTo (const T &val)
{
#ifndef RELEASE
  assert(buffer != 0);
#endif
  for (int i = 0; i < bufferSize; i++)
    buffer[i] = val;
}


template <class T>
void Image<T>::Resize (const int nbRow, const int nbCol)
{
#ifndef RELEASE
  assert((nbRow > 0) && (nbCol > 0));
#endif
  if ((nbRow*nbCol)!=(width*height)||(!deleteBuffer)) {
    DeAllocate ();
    Allocate (nbRow, nbCol);
  } else {
    height     = nbRow;
    width      = nbCol;
  }
}


template <class T>
void Image<T>::InitBuff(const T*const otherBuff, const int nbRow, const int nbCol)
{
#ifndef RELEASE
  assert((otherBuff != 0) && (nbRow > 0) && (nbCol > 0));
#endif
  DeAllocate ();
  Allocate (nbRow, nbCol);

  for (int i = 0; i < width * height; i++)
	  buffer[i] = otherBuff[i];
  
}


template <class T>
void Image<T>::ShallowSetBuff(T* const otherBuff, const int nbRow, const int nbCol)
{
#ifndef RELEASE
  assert((otherBuff != 0) && (nbRow > 0) && (nbCol > 0));
#endif
  DeAllocate ();
  height     = nbRow;
  width      = nbCol;
  bufferSize = height * width;
  buffer     = otherBuff;
}




