#ifndef REVERSIBLEWATERMARKING_HPP
#define REVERSIBLEWATERMARKING_HPP

#include <vector>

//Pour la fonction de sauvegarde des sites
#include "image.H"
#include "imageio.H"
#include "imagetools.H"


/**
 *  La classe est purement static.
 */
class ReversibleWatermarking {        

  enum SITESTATE {EMBEDDING, TO_CORRECT, ORIGINAL, UNDEFINE};

private:
        
  static bool IMPAIR(int x);
  static bool isPresent(int val, const vector<int>& lst);
  static int look_forward_for_next_emmbedding_site(int i, const unsigned char *c, int imsize, int n);
  static void saveImageSites( const vector<int>& embedding_sites, int height, int width);
  static void saveImageSites2( const Image<ReversibleWatermarking::SITESTATE>& imageSites);
  static void saveImageSites_Decodeur( const vector<int>& site_to_correct, int height, int width);
  static void checkSchemeReversibility ( const vector<int>& embedding_sites, const unsigned char *c, const unsigned char *cw, int height, int width, int n);
  static const int L; // Les valeurs des pixels appartiennent  l'interval [0 , L]
  static void get_message_and_correctingcode( const vector<int>& coefficients, // liste des coefficients extrait
                                              int nbSitesToCorrect,            // Nombre de sites  corriger        
                                              vector<int>& corecting_codes,    // liste des sites  corriger en step 2
                                              vector<int>& message,            // Suite de coefficients extrait compris dans [0, n-1]
                                              int n);                          // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
      
public:
 
//  static void embed( const unsigned char *c, // unsigned char *c, //image en entre
//                     int height,             // nombre de lignes
//                     int width,              // nombre de colonnes
//                     int* m,                 // Suite de coefficient compris dans [1, n]
//                     int mLen,               // Longueur du message 
//                     unsigned char *cw,      // Image marque
//                     int n);                 // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
//  static void embed2(const unsigned char *c, // unsigned char *c, //image en entre
//                     int height,             // nombre de lignes
//                     int width,              // nombre de colonnes
//                     int* m,                 // Suite binaire compose de 0 et de 1 ...
//                     int mLen,               // Longueur du message 
//                     unsigned char *cw,      // Image marque
//                     int n,                  // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
//                     vector<int>& corecting_codes);        
  static void embed3( const unsigned char *c,     // unsigned char *c, //image en entre
                      int height,                 // nombre de lignes
                      int width,                  // nombre de colonnes
                      const vector<int>& message, // Suite de coefficients extrait compris dans [0, n-1]
                      unsigned char *cw,          // Image marque
                      int n);                     // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
                                  
  static void embed_step1 ( const unsigned char *c,       // unsigned char *c, //image en entre
                            int height,                   // nombre de lignes
                            int width,                    // nombre de colonnes
                            unsigned char *cw,            // Image marque
                            vector<int>& corecting_codes, // Liste des code correcteur               
                            vector<int>& embedding_sites, // Liste des sites EMBEDDING
                            int n);                       // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
  static void embed_step2 ( unsigned char *cw,                  // Image marque
                            int height,                         // nombre de lignes
                            int width,                          // nombre de colonnes
                            const vector<int>& corecting_codes, // Liste des code correcteur               
                            const vector<int>& embedding_sites, // Liste des sites EMBEDDING
                            const vector<int>& message,         // Suite de coefficients extrait compris dans [0, n-1]
                            int n);                             // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
                            
  static void extract3( unsigned char *cw,    // image en entre
                        int height,           // height of image
                        int width,            // width of image
                        vector<int>& message, // Suite de coefficients extrait compris dans [0, n-1]
                        unsigned char *c,     // image rebuild
                        int n);               // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007

                                 
//  static void extract( unsigned char *cw,    // image en entre
//                       int height,           // height of image
//                       int width,            // width of image
//                       int* m,               // Suite de coefficient compris dans [1, n]
//                       int& mLen,            // longueur du message extrait
//                       unsigned char *c,     // image rebuild
//                       int n);               // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
//                       
//  static void extract2( unsigned char *cw,    // image en entre
//                        int height,           // height of image
//                        int width,            // width of image
//                        int* m,               // Suite de coefficient compris dans [1, n]
//                        int& mLen,            // longueur du message extrait
//                        unsigned char *c,     // image rebuild
//                        int n,               // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
//                        const vector<int>& corecting_codes);

  static void extract_step1 ( const unsigned char *cw,      // image en entre
                              int height,                   // height of image
                              int width,                    // width of image
                              unsigned char *c,             // image rebuild
                              vector<int>& coefficients,    // liste des coefficients extrait
                              vector<int>& site_to_correct, // liste des sites  corriger en step 2
                              int n);                       // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007
  static void extract_step2 ( const unsigned char *cw,            // image en entre
                              int height,                         // height of image
                              int width,                          // width of image
                              const vector<int>& coefficients,    // liste des coefficients extrait
                              const vector<int>& site_to_correct, // liste des sites  corriger en step 2
                              vector<int>& message,               // Suite de coefficients extrait compris dans [0, n-1]
                              unsigned char *c,                   // image rebuild
                              int n);                             // valeur suprieure ou gale  1 : cf. coltuc ICIP 2007

};    
#endif //REVERSIBLEWATERMARKING_HPP
